/** Exceptions thrown during lexing/parsing
*/
module antlrrt.exceptions;

private import antlrrt.utils;
private import antlrrt.lexing;
private import antlrrt.parsing;
private import antlrrt.collections;

public class RecognitionException : ANTLRException {
    public WString fileName;        // not used by treeparsers
    public int line;
    public int column;

    public this() {
        super(asWString("parsing error"));
        fileName = null;
        line = -1;
        column = -1;
    }

    /**
     * RecognitionException constructor comment.
     * @param s WString
     */
    public this(WString s) {
        super(s);
        fileName = null;
        line = -1;
        column = -1;
    }

    /** @deprecated As of ANTLR 2.7.2 use {@see #RecognitionException(wchar, WString, int, int) } */
    public this(WString s, WString fileName_, int line_) {
        this(s, fileName_, line_, -1);
    }
    
    /**
     * RecognitionException constructor comment.
     * @param s WString
     */
    public this(WString s, WString fileName_, int line_, int column_) {
        super(s);
        fileName = fileName_;
        line = line_;
        column = column_;
    }

    public WString getFilename() {
        return fileName;
    }

    public int getLine() {
        return line;
    }

    public int getColumn() {
        return column;
    }

    /** @deprecated As of ANTLR 2.7.0 */
    public WString getErrorMessage() {
        return getMessage();
    }

    public WString toWString() {
        return FileLineFormatter.getFormatter().
            getFormatString(fileName, line, column) ~ getMessage();
    }
}


public class NoViableAltException : RecognitionException {
    public Token token;
    public AST node;    // handles parsing and treeparsing

    public this(AST t) {
        super(asWString("NoViableAlt"), asWString("<AST>"), t.getLine(), t.getColumn());
        node = t;
    }

    public this(Token t, WString fileName_) {
        super(asWString("NoViableAlt"), fileName_, t.getLine(), t.getColumn());
        token = t;
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public WString getMessage() {
        if (token !is null) {
            return "unexpected token: " ~ token.getText();
        }

        // must a tree parser error if token is null
        if (node is TreeParser.ASTNULL) {
            return asWString("unexpected end of subtree");
        }
        return "unexpected AST node: " ~ node.toWString();
    }
}


public class NoViableAltForCharException : RecognitionException {
    public wchar foundChar;

    public this(wchar c, CharScanner scanner) {
        super(asWString("NoViableAlt"), scanner.getFilename(),
              scanner.getLine(), scanner.getColumn());
        foundChar = c;
    }

    /** @deprecated As of ANTLR 2.7.2 use {@see #NoViableAltForCharException(wchar, WString, int, int) } */
    public this(wchar c, WString fileName, int line) {
        this(c, fileName, line, -1);
    }

    public this(wchar c, WString fileName, int line, int column) {
        super(asWString("NoViableAlt"), fileName, line, column);
        foundChar = c;
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public WString getMessage() {
        WString mesg = asWString("unexpected char: ");

        // I'm trying to mirror a change in the C++ stuff.
        // But java seems to lack something convenient isprint-ish..
          // actually we're kludging around unicode and non unicode savy
          // output stuff like most terms.. Basically one would want to
          // be able to tweak the generation of this message.

        if ((foundChar >= ' ') && (foundChar <= '~')) {
            appS(mesg, '\'');
            appS(mesg, foundChar);
            appS(mesg, '\'');
        }
        else {
           appS(mesg, "0x" ~ asUpperHexWString(cast(int)foundChar)) ;   // +Integer.toHexString((int)foundChar).toUpperCase();
        }
        return mesg;
    }
}


public class SemanticException : RecognitionException {

    public this(WString s) {
        super(s);
    }

    /** @deprecated As of ANTLR 2.7.2 use {@see #SemanticException(wchar, WString, int, int) } */
    public this(WString s, WString fileName, int line) {
        this(s, fileName, line, -1);
    }

    public this(WString s, WString fileName, int line, int column) {
        super(s, fileName, line, column);
    }
}


public class MismatchedCharException : RecognitionException {
    // Types of chars
    public static const int CHAR = 1;
    public static const int NOT_CHAR = 2;
    public static const int RANGE = 3;
    public static const int NOT_RANGE = 4;
    public static const int SET = 5;
    public static const int NOT_SET = 6;

    // One of the above
    public int mismatchType;

    // what was found on the input stream
    public int foundChar;

    // For CHAR/NOT_CHAR and RANGE/NOT_RANGE
    public int expecting;

    // For RANGE/NOT_RANGE (expecting is lower bound of range)
    public int upper;

    // For SET/NOT_SET
    public BitSet set;

    // who knows...they may want to ask scanner questions
    public CharScanner scanner;

    /**
     * MismatchedCharException constructor comment.
     */
    public this() {
        super(asWString("Mismatched char"));
    }

    // Expected range / not range
    public this(wchar c, wchar lower, wchar upper_, boolean matchNot, CharScanner scanner_) {
        super(asWString("Mismatched char"), scanner_.getFilename(), scanner_.getLine(), scanner_.getColumn());
        mismatchType = matchNot ? NOT_RANGE : RANGE;
        foundChar = c;
        expecting = lower;
        upper = upper_;
        scanner = scanner_;
    }

    // Expected token / not token
    public this(wchar c, wchar expecting_, boolean matchNot, CharScanner scanner_) {
        super(asWString("Mismatched char"), scanner_.getFilename(), scanner_.getLine(), scanner_.getColumn());
        mismatchType = matchNot ? NOT_CHAR : CHAR;
        foundChar = c;
        expecting = expecting_;
        scanner = scanner_;
    }

    // Expected BitSet / not BitSet
    public this(wchar c, BitSet set_, boolean matchNot, CharScanner scanner_) {
        super(asWString("Mismatched char"), scanner_.getFilename(), scanner_.getLine(), scanner_.getColumn());
        mismatchType = matchNot ? NOT_SET : SET;
        foundChar = c;
        set = set_;
        scanner = scanner_;
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public WString getMessage() {
        StringBuffer sb = new StringBuffer();

        switch (mismatchType) {
            case CHAR:
                sb.append("expecting ");   appendCharName(sb, expecting);
                sb.append(", found ");     appendCharName(sb, foundChar);
                break;
            case NOT_CHAR:
                sb.append("expecting anything but '");
                appendCharName(sb, expecting);
                sb.append("'; got it anyway");
                break;
            case RANGE:
            case NOT_RANGE:
                sb.append("expecting token ");
                if (mismatchType == NOT_RANGE)
                    sb.append("NOT ");
                sb.append("in range: ");
                appendCharName(sb, expecting);
                sb.append("..");
                appendCharName(sb, upper);
                sb.append(", found ");
                appendCharName(sb, foundChar);
                break;
            case SET:
            case NOT_SET:
                sb.appendS("expecting " ~ (mismatchType == NOT_SET ? "NOT " : "") ~ "one of (");
                int[] elems = set.toArray();
                for (int i = 0; i < elems.length; i++) {
                    appendCharName(sb, elems[i]);
                }
                sb.append("), found ");
                appendCharName(sb, foundChar);
                break;
            default :
                sb.append(super.getMessage());
                break;
        }

        return sb.toWString();
    }

    /** Append a wchar to the msg buffer.  If special,
     *  then show escaped version
     */
    private void appendCharName(StringBuffer sb, int c) {
        switch (c) {
        case 65535 :
            // 65535 = cast(wchar) -1 = EOF
            sb.append("'<EOF>'");
            break;
        case '\n' :
            sb.append("'\\n'");
            break;
        case '\r' :
            sb.append("'\\r'");
            break;
        case '\t' :
            sb.append("'\\t'");
            break;
        default :
            sb.append('\'');
            sb.append(cast(wchar) c);
            sb.append('\'');
            break;
        }
    }
}



public class MismatchedTokenException : RecognitionException {
    // Token names array for formatting
    WString[] tokenNames;
    // The token that was encountered
    public Token token;
    // The offending AST node if tree walking
    public AST node;

    WString tokenText = null; // taken from node or token object

    // Types of tokens
    public static const int TOKEN = 1;
    public static const int NOT_TOKEN = 2;
    public static const int RANGE = 3;
    public static const int NOT_RANGE = 4;
    public static const int SET = 5;
    public static const int NOT_SET = 6;
    // One of the above
    public int mismatchType;

    // For TOKEN/NOT_TOKEN and RANGE/NOT_RANGE
    public int expecting;

    // For RANGE/NOT_RANGE (expecting is lower bound of range)
    public int upper;

    // For SET/NOT_SET
    public BitSet set;

    /** Looking for AST wildcard, didn't find it */
    public this() {
        super(asWString("Mismatched Token: expecting any AST node"), asWString("<AST>"), -1, -1);
    }

    // Expected range / not range
    public this(WString[] tokenNames_, AST node_, int lower, int upper_, boolean matchNot) {
        super(asWString("Mismatched Token"), asWString("<AST>"), node_ is null? -1:node_.getLine(), node_ is null? -1:node_.getColumn());
        tokenNames = tokenNames_;
        node = node_;
        if (node_ is null) {
            tokenText = asWString("<empty tree>");
        }
        else {
            tokenText = node_.toWString();
        }
        mismatchType = matchNot ? NOT_RANGE : RANGE;
        expecting = lower;
        upper = upper_;
    }

    // Expected token / not token
    public this(WString[] tokenNames_, AST node_, int expecting_, boolean matchNot) {
        super(asWString("Mismatched Token"), asWString("<AST>"), node_ is null? -1:node_.getLine(), node_ is null? -1:node_.getColumn());
        tokenNames = tokenNames_;
        node = node_;
        if (node_ is null) {
            tokenText = asWString("<empty tree>");
        }
        else {
            tokenText = node_.toWString();
        }
        mismatchType = matchNot ? NOT_TOKEN : TOKEN;
        expecting = expecting_;
    }

    // Expected BitSet / not BitSet
    public this(WString[] tokenNames_, AST node_, BitSet set_, boolean matchNot) {
        super(asWString("Mismatched Token"), asWString("<AST>"), node_ is null? -1:node_.getLine(), node_ is null? -1:node_.getColumn());
        tokenNames = tokenNames_;
        node = node_;
        if (node_ is null) {
            tokenText = asWString("<empty tree>");
        }
        else {
            tokenText = node_.toWString();
        }
        mismatchType = matchNot ? NOT_SET : SET;
        set = set_;
    }

    // Expected range / not range
    public this(WString[] tokenNames_, Token token_, int lower, int upper_, boolean matchNot, WString fileName_) {
        super(asWString("Mismatched Token"), fileName_, token_.getLine(), token_.getColumn());
        tokenNames = tokenNames_;
        token = token_;
        tokenText = token_.getText();
        mismatchType = matchNot ? NOT_RANGE : RANGE;
        expecting = lower;
        upper = upper_;
    }

    // Expected token / not token
    public this(WString[] tokenNames_, Token token_, int expecting_, boolean matchNot, WString fileName_) {
        super(asWString("Mismatched Token"), fileName_, token_.getLine(), token_.getColumn());
        tokenNames = tokenNames_;
        token = token_;
        tokenText = token_.getText();
        mismatchType = matchNot ? NOT_TOKEN : TOKEN;
        expecting = expecting_;
    }

    // Expected BitSet / not BitSet
    public this(WString[] tokenNames_, Token token_, BitSet set_, boolean matchNot, WString fileName_) {
        super(asWString("Mismatched Token"), fileName_, token_.getLine(), token_.getColumn());
        tokenNames = tokenNames_;
        token = token_;
        tokenText = token_.getText();
        mismatchType = matchNot ? NOT_SET : SET;
        set = set_;
    }

    /**
     * Returns a clean error message (no line number/column information)
     */
    public WString getMessage() {
        StringBuffer sb = new StringBuffer();

        switch (mismatchType) {
            case TOKEN:
                sb.append("expecting " ~ tokenName(expecting) ~ ", found '" ~ tokenText ~ "'");
                break;
            case NOT_TOKEN:
                sb.append("expecting anything but " ~ tokenName(expecting) ~ "; got it anyway");
                break;
            case RANGE:
                sb.append("expecting token in range: " ~ tokenName(expecting) ~ ".." ~ tokenName(upper) ~ ", found '" ~ tokenText ~ "'");
                break;
            case NOT_RANGE:
                sb.append("expecting token NOT in range: " ~ tokenName(expecting) ~ ".." ~ tokenName(upper) ~ ", found '" ~ tokenText ~ "'");
                break;
            case SET:
            case NOT_SET:
                sb.appendS("expecting " ~ (mismatchType == NOT_SET ? "NOT " : "") ~ "one of (");
                int[] elems = set.toArray();
                for (int i = 0; i < elems.length; i++) {
                    sb.append(" ");
                    sb.append(tokenName(elems[i]));
                }
                sb.append("), found '" ~ tokenText ~ "'");
                break;
            default :
                sb.append(super.getMessage());
                break;
        }

        return sb.toWString();
    }

    private WString tokenName(int tokenType) {
        if (tokenType == Token.INVALID_TYPE) {
            return asWString("<Set of tokens>");
        }
        else if (tokenType < 0 || tokenType >= tokenNames.length) {
            return "<" ~ asWString(tokenType) ~ ">";
        }
        else {
            return tokenNames[tokenType];
        }
    }
}


public class TokenStreamException : ANTLRException {
    public this() {
    }

    public this(WString s) {
        super(s);
    }
}


/**
 * Wraps an IOException in a TokenStreamException
 */
public class TokenStreamIOException : TokenStreamException {
    public IOException io;

    /**
     * TokenStreamIOException constructor comment.
     * @param s WString
     */
    public this(IOException io) {
        super(asWString(io.toString()));
        this.io = io;
    }
}


/**
 * Wraps a RecognitionException in a TokenStreamException so you
 * can pass it along.
 */
public class TokenStreamRecognitionException : TokenStreamException {
    public RecognitionException recog;

    public this(RecognitionException re) {
        super(re.getMessage());
        this.recog = re;
    }

    public WString toWString() {
        return recog.toWString();
    }
}


/**
 * Aborted recognition of current token. Try to get one again.
 * Used by TokenStreamSelector.retry() to force nextToken()
 * of stream to re-enter and retry.
 */
public class TokenStreamRetryException : TokenStreamException {
    public this() {
    }
}


/**
 * Anything that goes wrong while generating a stream of characters
 */
public class CharStreamException : ANTLRException {
    /**
     * CharStreamException constructor comment.
     * @param s WString
     */
    public this(WString s) {
        super(s);
    }
}


/**
 * Wrap an IOException in a CharStreamException
 */
public class CharStreamIOException : CharStreamException {
    public IOException io;

    public this(IOException io) {
        super(asWString(io.toString()));
        this.io = io;
    }
}






